@@ -3,6 +3,7 @@ from abc import ABC, abstractmethod
 from contextlib import contextmanager
 from functools import partial
 from .wadray import Wad, Ray
+import requests
 from environs import Env
 
 env = Env()
@@ -10,6 +11,10 @@ env = Env()
 SKIP_PROXY = env.bool("SKIP_PROXY", False)
 DEFAULT_PROVIDER = env.str("DEFAULT_PROVIDER", None)
 
+ETHERSCAN_TOKEN = env.str("ETHERSCAN_TOKEN", None)
+ETHERSCAN_DOMAIN = env.str("ETHERSCAN_DOMAIN", "api.etherscan.io")
+ETHERSCAN_URL = env.str("ETHERSCAN_URL", "https://{domain}/api?apikey={token}&")
+
 MAXUINT256 = 2**256 - 1
 
 _providers = {}
@@ -89,6 +94,10 @@ class AddressBook(ABC):
     def set_instance(cls, obj):
         cls.instance = obj
 
+    def get_signer_account(self, address):
+        """Returns a LocalAccount or other object that can sign transactions"""
+        raise NotImplementedError()
+
 
 class ETHCall(ABC):
     def __init__(self, eth_method, eth_args, eth_return_type="", adapt_args=None, eth_variant=None):
@@ -98,10 +107,21 @@ class ETHCall(ABC):
         self.adapt_args = adapt_args
         self.eth_variant = eth_variant
 
-    @abstractmethod
-    def _get_eth_function(self, wrapper, eth_method, eth_variant=None):
+    @classmethod
+    def get_eth_function(cls, wrapper, eth_method, eth_variant=None):
+        return cls.get_eth_function_and_mutability(wrapper, eth_method, eth_variant)[0]
+
+    @classmethod
+    def get_eth_function_and_mutability(cls, wrapper, eth_method, eth_variant=None):
         raise NotImplementedError()
 
+    def normalize_receipt(self, wrapper, receipt):
+        """
+        Function to normalize receipts to behave somewhat similar
+        (taking brownie receipts as interface for now)
+        """
+        return receipt
+
     def _handle_exception(self, err):
         raise err
 
@@ -130,12 +150,16 @@ class ETHCall(ABC):
             msg_args["from"] = wrapper._auto_from
         call_args.append(msg_args)
 
-        eth_function = self._get_eth_function(wrapper, self.eth_method, self.eth_variant)
+        eth_function, mutability = self.get_eth_function_and_mutability(
+            wrapper, self.eth_method, self.eth_variant
+        )
 
         try:
             ret_value = eth_function(*call_args)
         except Exception as err:
             self._handle_exception(err)
+        if mutability in ("payable", "nonpayable"):  # ret_value is a receipt
+            ret_value = self.normalize_receipt(wrapper, ret_value)
         return self.unparse(wrapper, self.eth_return_type, ret_value)
 
     @classmethod
@@ -169,15 +193,70 @@ class ETHCall(ABC):
 
     @classmethod
     def unparse(cls, wrapper, value_type, value):
+        if value_type.startswith("(") and value_type.endswith(")"):
+            # It's a tuple / struct
+            value_types = [t.strip() for t in value_type.strip("()").split(",")]
+            return tuple(
+                cls.unparse(wrapper, vt, value[i]) for i, vt in enumerate(value_types)
+            )
         if value_type == "amount":
             return Wad(value)
         if value_type == "ray":
             return Ray(value)
         if value_type == "address":
-            return wrapper.provider.address_book.get_name(value)
+            name = wrapper.provider.address_book.get_name(value)
+            return name or value
+        if value_type == "":
+            wrapper.add_receipt(value)
+            return value
+        if value_type == "receipt":
+            wrapper.add_receipt(value)
+            return value
         return value
 
 
+class BaseProvider(ABC):
+    @abstractmethod
+    def get_contract_factory(self, eth_contract):
+        raise NotImplementedError()
+
+    @abstractmethod
+    def get_events(self, eth_wrapper, event_name, filter_kwargs={}):
+        raise NotImplementedError()
+
+    def get_etherscan_url(self):
+        if ETHERSCAN_TOKEN is None:
+            return None
+        return ETHERSCAN_URL.format(token=ETHERSCAN_TOKEN, domain=ETHERSCAN_DOMAIN)
+
+    def get_first_block(self, eth_wrapper):
+        etherscan_url = self.get_etherscan_url()
+        if not etherscan_url:
+            return 0
+        address = self.get_contract_address(eth_wrapper)
+        url = (
+            etherscan_url + f"&module=account&action=txlist&address={address}&startblock=0&" +
+            "endblock=99999999&page=1&offset=10&sort=asc"
+        )
+        resp = requests.get(url)
+        resp.raise_for_status()
+        resp = resp.json()
+        if not resp["result"]:
+            return -1
+        return int(resp["result"][0]["blockNumber"])
+
+    def get_contract_address(self, eth_wrapper):
+        return eth_wrapper.contract.address
+
+    @abstractmethod
+    def init_eth_wrapper(self, eth_wrapper, owner, init_params, kwargs):
+        pass
+
+    @abstractmethod
+    def build_contract(self, contract_address, contract_factory, contract_name=None):
+        raise NotImplementedError
+
+
 class ETHWrapper:
     proxy_kind = None
     libraries_required = []
@@ -198,6 +277,17 @@ class ETHWrapper:
     def provider(self):
         return get_provider(self.provider_key)
 
+    def add_receipt(self, receipt):
+        if not hasattr(self, "_receipts"):
+            self._receipts = []
+        self._receipts.append(receipt)
+
+    @property
+    def last_receipt(self):
+        if not hasattr(self, "_receipts") or not self._receipts:
+            return None
+        return self._receipts[-1]
+
     @property
     def eth_call(self):
         return self.provider.eth_call
@@ -231,6 +321,11 @@ class ETHWrapper:
         return self.provider.address_book.get_name(account)
 
     grant_role = MethodAdapter((("role", "keccak256"), ("user", "address")))
+    revoke_role = MethodAdapter((("role", "keccak256"), ("user", "address")))
+    renounce_role = MethodAdapter((("role", "keccak256"), ("user", "address")))
+    has_role = MethodAdapter((("role", "keccak256"), ("user", "address")), "bool")
+    get_role_admin = MethodAdapter((("role", "keccak256"), ), "address")
+    get_role_admin = MethodAdapter((("role", "keccak256"), ), "bytes32")
 
     @contextmanager
     def as_(self, user):
@@ -244,6 +339,9 @@ class ETHWrapper:
             else:
                 self._auto_from = prev_auto_from
 
+    def set_auto_from(self, user):
+        self._auto_from = user
+
 
 class IERC20(ETHWrapper):
     eth_contract = "IERC20Metadata"
@@ -255,11 +353,11 @@ class IERC20(ETHWrapper):
     balance_of = MethodAdapter((("account", "address"), ), "amount")
     transfer = MethodAdapter((
         ("sender", "msg.sender"), ("recipient", "address"), ("amount", "amount")
-    ), "bool")
+    ), "receipt")
 
     allowance = MethodAdapter((("owner", "address"), ("spender", "address")), "amount")
     approve = MethodAdapter((("owner", "msg.sender"), ("spender", "address"), ("amount", "amount")),
-                            "bool")
+                            "receipt")
     increase_allowance = MethodAdapter(
         (("owner", "msg.sender"), ("spender", "address"), ("amount", "amount"))
     )
@@ -269,7 +367,7 @@ class IERC20(ETHWrapper):
 
     transfer_from = MethodAdapter((
         ("spender", "msg.sender"), ("sender", "address"), ("recipient", "address"), ("amount", "amount")
-    ), "bool")
+    ), "receipt")
 
 
 class IERC721(ETHWrapper):
@@ -280,7 +378,7 @@ class IERC721(ETHWrapper):
     owner_of = MethodAdapter((("token_id", "int"), ), "address")
     approve = MethodAdapter((
         ("sender", "msg.sender"), ("spender", "address"), ("token_id", "int")
-    ), "bool")
+    ), "receipt")
     get_approved = MethodAdapter((("token_id", "int"), ), "address")
     set_approval_for_all = MethodAdapter((
         ("sender", "msg.sender"), ("operator", "address"), ("approved", "bool")
@@ -288,7 +386,7 @@ class IERC721(ETHWrapper):
     is_approved_for_all = MethodAdapter((("owner", "address"), ("operator", "address")), "bool")
     transfer_from = MethodAdapter((
         ("spender", "msg.sender"), ("from", "address"), ("to", "address"), ("token_id", "int")
-    ), "bool")
+    ), "receipt")
     transfer = MethodAdapter((
         ("sender", "msg.sender"), ("recipient", "address"), ("amount", "amount")
-    ), "bool")
+    ), "receipt")
